;*=====================================================================*/
;*    serrano/prgm/project/bigloo/fthread/src/Llib/scheduler.scm       */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Wed Feb  6 14:39:43 2002                          */
;*    Last change :  Fri Jun 13 10:39:51 2003 (serrano)                */
;*    Copyright   :  2002-03 Manuel Serrano                            */
;*    -------------------------------------------------------------    */
;*    Fair threads scheduler                                           */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __ft_scheduler
   
   (import __ft_types
	   __ft_%types
	   __ft_%scheduler
	   __ft_%thread
	   __ft_thread
	   __ft_env
	   __ft_%env
	   __ft_signal
	   __ft_exception)
   
   (export (make-scheduler . env)
	   (current-scheduler)
	   (default-scheduler . scdl)
	   (with-scheduler ::scheduler ::procedure)
	   (scheduler-react! . ::obj)
	   (scheduler-start! . args)
	   (scheduler-terminate! . ::obj)
	   (scheduler-instant::int . ::obj)
	   (scheduler-broadcast! ::scheduler event . val)
	   
	   ;; misc 
	   (broadcast! event . val)))
	   
;*---------------------------------------------------------------------*/
;*    *current-scheduler* & *default-scheduler* ...                    */
;*---------------------------------------------------------------------*/
(define *current-scheduler* #f)
(define *default-scheduler* #f)

;*---------------------------------------------------------------------*/
;*    current-scheduler ...                                            */
;*---------------------------------------------------------------------*/
(define (current-scheduler) *current-scheduler*)

;*---------------------------------------------------------------------*/
;*    default-scheduler ...                                            */
;*---------------------------------------------------------------------*/
(define (default-scheduler . scdl)
   (cond
      ((null? scdl)
       *default-scheduler*)
      ((scheduler? (car scdl))
       (set! *default-scheduler* (car scdl))
       (car scdl))
      (else
       (error "default-scheduler" "Illegal scheduler" (car scdl)))))

;*---------------------------------------------------------------------*/
;*    with-scheduler ...                                               */
;*---------------------------------------------------------------------*/
(define (with-scheduler s thunk)
   (let ((old (default-scheduler)))
      (unwind-protect
	 (begin
	    (default-scheduler s)
	    (thunk))
	 (default-scheduler old))))

;*---------------------------------------------------------------------*/
;*    make-scheduler ...                                               */
;*---------------------------------------------------------------------*/
(define (make-scheduler . envs)
   (let ((the-sched-name (gensym 'scheduler)))
      (letrec ((s (instantiate::%scheduler
		     (%builtin (%bgl-thread-new/name
				(lambda () (schedule s))
				the-sched-name))
		     (name the-sched-name)
		     (env (append envs (list (instantiate::%env))))
		     (%exc-handlers (list (lambda (x) x))))))
	 ;; initialize the back pointer facility
	 (thread-setup! s)
	 ;; start the native thread (that will block instantly)
	 (%bgl-thread-start (thread-%builtin s))
	 ;; if there is no default scheduler, store that one
	 (if (not (default-scheduler)) (default-scheduler s))
	 ;; return the newly allocates scheduler
	 s)))

;*---------------------------------------------------------------------*/
;*    scheduler-react! ...                                             */
;*---------------------------------------------------------------------*/
(define (scheduler-react! . s)
   (let ((s (cond
	       ((null? s)
		(default-scheduler))
	       ((scheduler? (car s))
		(car s))
	       (else
		(error "scheduler-react!"
		       "Illegal scheduler"
		       (car s)))))
	 (o (current-scheduler)))
      (with-access::%scheduler s (tostart
				  deadn startn
				  toterminate toresume
				  threads-yield threads-eoi)
	 (set! *current-scheduler* s)
	 (if (> *thread-debug* 3) (print "Switch in scheduler " s))
	 (thread-switch-in s)
	 (if (> *thread-debug* 3) (print "Came back from scheduler " s))
	 (set! *current-scheduler* o)
	 (if (> *thread-debug* 1) (print "deadn=" deadn " startn=" startn))
	 (cond
	    ((or (pair? tostart) (pair? toresume))
	     'ready)
	    ((=fx deadn startn)
	     'done)
	    ((>=fx (+fx deadn (length toterminate)) startn)
	     'ready)
	    ((pair? (%bgl-asynchronous-thunk-scheduler-funcall
		     (lambda (s) (%scheduler-tobroadcast s))
		     s))
	     'ready)
	    ((pair? threads-eoi)
	     'ready)
	    ((null? threads-yield)
	     'await)
	    (else
	     'ready)))))

;*---------------------------------------------------------------------*/
;*    scheduler-react-debug ...                                        */
;*---------------------------------------------------------------------*/
(define (scheduler-react-debug s res)
   ;; a simple debugging function used to display the result
   ;; of each reaction
   (print "scheduler-react!: " res)
   (if (and (> *thread-debug* 1)
	    (eq? res 'await)
	    (%env? (car (scheduler-env s))))
       (begin
	  (print "           await: ")
	  (for-each (lambda (sgn)
		       (print "               -- "
			      (%signal-id sgn) ": ")
		       (print "                  "
			      (%signal-threads sgn))
		       (newline))
		    (%env-waiting-signals
		     (car (scheduler-env s))))
	  (newline))))

;*---------------------------------------------------------------------*/
;*    scheduler-start! ...                                             */
;*---------------------------------------------------------------------*/
(define (scheduler-start! . args)
   (let ((iterp (and (pair? args) (number? (car args))))
	 (next (cond
		  ((null? args)
		   (lambda (i)
		      (+fx i 1)))
		  ((number? (car args))
		   (lambda (i)
		      (if (< i (car args))
			  (+fx i 1)
			  #f)))
		  ((procedure? (car args))
		   (lambda (i)
		      (let ((res ((car args) i)))
			 (if res (+fx i 1) #f))))
		  (else
		   (error "scheduler-start!"
			  "Illegal optional parameter"
			  args))))
	 (s (cond
	       ((or (null? args) (null? (cdr args)))
		(default-scheduler))
	       ((scheduler? (cadr args))
		(cadr args))
	       (else
		(error "scheduler-start!" "Illegal scheduler" (cadr args))))))
      (let loop ((i 1))
	 (if i
	     (let ((res (scheduler-react! s)))
		(if (> *thread-debug* 0) (scheduler-react-debug s res))
		(case res
		   ((await)
		    (if (not iterp)
			(%bgl-asynchronous-thunk-scheduler-funcall
			 (lambda (s)
			    (begin
			       (if (not (pair? (%scheduler-tobroadcast s)))
				   (begin
				      (%bgl-asynchronous-thunk-await! s)
				      #unspecified))))
			 s))
		    (loop (next i)))
		   ((ready)
		    (loop (next i)))
		   (else
		    res)))))))

;*---------------------------------------------------------------------*/
;*    scheduler-terminate! ...                                         */
;*---------------------------------------------------------------------*/
(define (scheduler-terminate! . s)
   (let ((s (cond
	       ((null? s)
		(default-scheduler))
	       ((scheduler? (car s))
		(car s))
	       (else
		(error "scheduler-react!"
		       "Illegal scheduler"
		       (car s))))))
      (with-access::%scheduler s (tosuspend toresume tostart threads)
	 (set! toresume '())
	 (set! tosuspend '())
	 (set! tostart '())
	 (for-each thread-terminate! threads))))

;*---------------------------------------------------------------------*/
;*    scheduler-broadcast! ...                                         */
;*---------------------------------------------------------------------*/
(define (scheduler-broadcast! s event . val)
   (%bgl-asynchronous-thunk-scheduler-funcall2
    (lambda (s evt)
       (with-access::%scheduler s (tobroadcast)
	  (set! tobroadcast (cons evt tobroadcast))))
    s
    (cons event (if (pair? val) (car val) #unspecified))))

;*---------------------------------------------------------------------*/
;*    scheduler-instant ...                                            */
;*---------------------------------------------------------------------*/
(define (scheduler-instant . s)
   (let ((s (cond
	       ((null? s)
		(default-scheduler))
	       ((scheduler? (car s))
		(car s))
	       (else
		(error "scheduler-react!" "Illegal scheduler" (car s))))))
      (with-access::%scheduler s (env)
	 (ftenv-instant (car env)))))

;*---------------------------------------------------------------------*/
;*    broadcast! ...                                                   */
;*---------------------------------------------------------------------*/
(define (broadcast! event . val)
   (let ((t (current-thread)))
      (if (thread? t)
	  (if (thread-attached? t)
	      (with-access::thread t (scheduler)
		 (with-access::%scheduler scheduler (env)
		    (signal-emit event
				 (if (pair? val)
				     (car val)
				     #unspecified)
				 env)))
	      (error "broadcast!" "Unattached thread" t)))))

;*---------------------------------------------------------------------*/
;*    sort-threads ...                                                 */
;*---------------------------------------------------------------------*/
(define (sort-threads threads)
   (if *thread-strict-order*
       (sort threads
	     (lambda (t1 t2)
		(<fx (thread-%ident t1)
		     (thread-%ident t2))))
       threads))

;*---------------------------------------------------------------------*/
;*    schedule ...                                                     */
;*---------------------------------------------------------------------*/
(define (schedule scdl::scheduler)
   (define (init-instant)
      (with-access::%scheduler scdl (threads-runnable threads-yield env eoi)
	 ;; prepare the environment
	 (for-each instant-env! env)
	 ;; prepare the scheduler
	 (set! eoi #f)
	 (set! threads-runnable threads-yield)
	 (set! threads-yield '())))
   (define (debug-run threads i state)
      ;; dump thread (this is a debug function)
      (if (> *thread-debug* 0)
	  (with-output-to-port (current-error-port)
	     (lambda ()
		(let ((m (make-string 30 #\space)))
		   (print m "--- " state ":")
		   (if (> *thread-debug* 1)
		       (for-each (lambda (t)
				    (print m "  " t 
					   " [" (thread-%state t) ":"
					   (thread-%timeout t) "]"))
				 threads)
		       (print m "thread candidates: "  (length threads)))
		   (print m "------------------------------------ "))))))
   (define (thread-runnable? t)
      (with-access::thread t (%timeout %state)
	 (cond
	    ((eq? %state 'suspended)
	     ;; the thread is suspended, do nothing
	     #f)
	    ((<=fx %timeout 1)
	     #t)
	    ((eq? %state 'yield)
	     ;; the thread is coooperating
	     (thread-set-yield! t (-fx %timeout 1))
	     #f)
	    ((eq? %state 'await)
	     ;; the thread is waiting with a timeout
	     (thread-set-eoi! t (-fx %timeout 1))
	     #f)
	    (else
	     #f))))
   (define (run-threads threads state)
      ;; run all the threads
      (if (pair? threads)
	  (with-access::%scheduler scdl (torun)
	     (let* ((t* (filter! thread-runnable? threads))
		    (ts* (sort-threads t*)))
		(debug-run ts* (ftenv-instant (car (scheduler-env scdl))) state)
		(if (pair? ts*)
		    (let ((t (car ts*)))
		       (set! torun (cdr ts*))
		       (%scheduler-current-thread-set! scdl t)
		       (thread-switch-to scdl t)))))))
   (define (start-threads)
      (with-access::%scheduler scdl (tostart threads-runnable)
	 (if (pair? tostart)
	     (begin
		;; append the new threads at the end of the threads list
		(set! threads-runnable
		      (append! (reverse! tostart) threads-runnable))
		(set! tostart '())))))
   (define (terminate-threads)
      (with-access::%scheduler scdl (toterminate threads-runnable)
	 (if (pair? toterminate)
	     (begin
		;; we have to remove multiple occurrences of thread to be
		;; terminated otherwise we will have a lock because we will run
		;; a terminated thread
		(let ((thds (filter (lambda (t)
				       (with-access::thread t (%state %timeout)
					  (if (not (memq %state '(dying dead)))
					      (begin
						 (set! %state 'dying)
						 (set! %timeout 0)
						 #t)
					      #f)))
				    toterminate)))
		   (set! toterminate '())
		   ;; mark that the thread is dying and remove it from the
		   ;; runnable and await list
		   (for-each (lambda (t)
				(set! threads-runnable
				      (remq! t threads-runnable)))
			     thds)
		   (run-threads thds 'dying))))))
   (define (suspend-threads)
      (with-access::%scheduler scdl (tosuspend threads-suspend)
	 (if (pair? tosuspend)
	     (begin
		(for-each (lambda (t)
			     (thread-%state-set! t 'suspended))
			  tosuspend)
		(set! threads-suspend (append! tosuspend threads-suspend))
		(set! tosuspend '())))))
   (define (resume-threads)
      (with-access::%scheduler scdl (toresume threads-suspend threads-runnable)
	 (if (pair? toresume)
	     (begin
		;; append the new threads at the end of the threads list
		(for-each (lambda (t)
			     (thread-%state-set! t 'ready)
			     (set! threads-suspend (remq! t threads-suspend))
			     (set! threads-runnable (cons t threads-runnable)))
			  toresume)
		(set! toresume '())))))
   (define (broadcast-events)
      (if (> *thread-debug* 0)
	  (fprint (current-error-port)
		  "                              scheduler-broadcast:"))
      (%bgl-asynchronous-thunk-scheduler-funcall
       (lambda (s)
	  (with-access::%scheduler s (tobroadcast env)
	     ;; generate all the signals
	     (for-each (lambda (x)
			  (if (> *thread-debug* 0)
			      (fprint (current-error-port)
				      "                                + "
				      (car x) " " (cdr x)))
			  (signal-emit (car x) (cdr x) env))
		       tobroadcast)
	     (set! tobroadcast '())))
       scdl)
      (if (> *thread-debug* 0)
	  (print
	   (make-string 30 #\space) "------------------------------------ ")))
   (define (start-lazy-signals)
      (with-access::%scheduler scdl (signals-tostart)
	 (if (> *thread-debug* 5)
	     (print "Start waiting signals " signals-tostart))
	 (for-each (lambda (e)
		      (with-access::%lazy-signal e (run action scheduler id)
			 (if (not run)
			     (begin
				(set! run #t)
				(let ((a (%bgl-asynchronous-thunk-new action
								      scheduler
								      e
								      id)))
				   (%bgl-asynchronous-thunk-start a)
				   #unspecified)))))
		   signals-tostart)
	 (set! signals-tostart '())))
   (define (execute-threads)
      (with-access::%scheduler scdl (current-thread
				     env eoi
				     threads-runnable
				     threads-eoi)
	 (let loop ()
	    ;; the runnable threads
	    (let ((ths threads-runnable))
	       (set! threads-runnable '())
	       (run-threads ths 'runnable))
	    ;; we check if there still are runnable threads
	    (if (pair? threads-runnable)
		(loop)
		(begin
		   (set! current-thread #f)
		   (set! eoi #t)
		   (let ((ths threads-eoi))
		      (set! threads-eoi '())
		      (run-threads ths 'eoi)))))))
   ;; prepare the threads and the instant numbering
   (init-instant)
   (if (> *thread-debug* 3) (print "schedule: " scdl))
   (if (> *thread-debug* 0)
       (print "\n" (make-string 29 #\=) " instant: "
	      (ftenv-instant (car (scheduler-env scdl)))
	      " ========================="))
   ;; add all the new threads
   (start-threads)
   ;; resume suspended threads
   (resume-threads)
   ;; suspend threads
   (suspend-threads)
   ;; prepare the threads that are about to die
   (terminate-threads)
   ;; broadcast the events
   (broadcast-events)
   ;; run until the end of the install
   (execute-threads)
   ;; start lazy signals
   (start-lazy-signals)
   ;; leave the thread implementing this scheduler
   (thread-switch-out scdl)
   ;; when the scheduler will get the CPU, it will have to loop,
   ;; that is, to restart its attached threads.
   (schedule scdl))

